
/*** rcal.c - Regression Calibration */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "qvf.h"
#include "rcal.h"

/*
 *   The rcal procedure:
 */

/*** Calculate covariance using internal reps. NOTE: calcw() must must
     be called before calcSuu(), d->x must contain averaged reps (\bar
     W). This code assumes that we will have the same number of reps for
     each obervation/subject. It does not handle the completly generic
     case of different replications for different variables and subjects
     (eg. in d->[i][j] we must have the same number of reps for each
     i). We do not check that this is case in calcsuu(). *ws is a work 
     vector of size p. */

int calcSuu (dclData *d, dclQvf *q, double **suu, double *ws)
{
  int r, c, i, j, divisor, reps;

  dmatrixZero (suu, d->px, d->px);

  divisor = 0;
  for (i=1; i<=d->n; ++i) {
    reps = d->kn[i][1];  

    for (j=1; j<=reps; ++j) {
      for (r=1; r<=d->px; ++r) {
/*r is the variable, i is the subject, j is the rep*/
        ws[r] = d->W[i][((r-1)*reps)+j] - d->x[i][r];
      }

      for (r=1; r<=d->px; ++r) {
        for (c=1; c<=r; ++c) {
          suu[r][c] += ws[r] * ws[c];
        }
      }
    }
    divisor += (d->kn[i][1] - 1);
  }

  if (divisor > 0) {
    for (r=1; r<=d->px; ++r) {
      for (c=1; c<=r; ++c) {
        suu[r][c] /= (double) divisor;
        suu[c][r] = suu[r][c];
      }
    }
  }

  /*** If the divisor is zero, then we have no replications in our
       measurement error data. We exit or return with error (in case this  
       happened during bootstraping. */

  else {
    if (!q->booting) 
      error (StatusNoReplications,
             "No replications in measurement error data. Specify suu.");
    return (1);
  }

  

  return (0);
}


/*** Calculate average'd X (averaged measurement error data) matrix */

void calcWbar (dclData *d)
{
  int i, j, k, col;

  for (i=1; i <= d->n; ++i) {
    col = 1;
    for (j=1; j <= d->px; ++j) {
      d->x[i][j] = d->W[i][col++];
      for (k=1; k < d->kn[i][j]; ++k)
        d->x[i][j] += d->W[i][col++];
      d->x[i][j] /= (double) d->kn[i][j];
    }
  }
  
}


/*** Regression Calibration. */

void rcal (dclData *d, dclQvf *q, dclRcal *r)  
{
  int i, j, k, err, kmax, reps;
  int n, px, pz, **kn;
  double c, v, vk, vk2;
  double **z, **ww, *zm, *wm, **sxx, **sxz, **szz, **suu, 
         **s4, *wz, *wi, **ex, ***S4;
  char *suutype;


  note (6, "*** Rcal: Calculating E[X] from W");

  /*** these matrices are column x rows !!! */

  z   = d->z;   /* n*pz; z matrix - observations with no error */
  ww  = d->x;   /* n*px; \bar w matrix (averaged reps) - obs. with error */
  ex  = r->ex;  /* n*px; Output matrix - new W matrix */


  /*** regular vector & matrices */

  n   = d->n;   /* number of observations without reps */
  px  = d->px;  /* number of columns in w matrix */
  pz  = d->pz;  /* number of columns in z matrix */
  kn  = d->kn;  /* px*n; matrix contain # of reps per obervation */
  zm  = r->zm;  /* pz; column average of Z matrix */
  wm  = r->wm;  /* px; column average of W matrix */
  sxx = r->sxx; /* px*px; */
  sxz = r->sxz; /* px*pz; */
  szz = r->szz; /* pz*pz; */
  suu = r->suu; /* px*px; */

  wz  = dvector (px+pz);  
  wi  = dvector (px+pz);


  /*** We calculate \bar W */

  calcWbar (d);
  mess (5, dmatrixPrint (d->x, d->n, d->px, "w bar"));

  /*** Then we calculate the covariance matrix using internal
       replications (unless the user supplied the matrix). */

  if (r->uutype != CovType1) { 
    err = calcSuu(d, q, r->suu, r->wm); 
    if (err) {
      q->status = StatusNoReplications;    
      return; 
    }
    suutype = "Suu (by replications)";
  } 
  else {
    suutype = "Suu (user supplied)";
  }
  mess (5, dmatrixPrint (r->suu, d->px, d->px, suutype));

  /*** calculate v, nu, replicate related numbers */

  kmax = 0;
  vk = vk2 = 0.0;
  for (i=1; i<=n; ++i) {
    k = kn[i][1];
    vk  += k;
    vk2 += k*k;
    if (k > kmax) kmax=k;
  }
  v    = vk - (vk2/vk);
  r->v = v; 
  note (6, "sum k = %f, sum k^2 = %f, nu = %f, kmax=%d", vk, vk2, v, kmax);

  /*** calculate Z & W mean vectors */

  for (j=1; j<=pz; ++j) { 
    zm[j] = 0.0;
    for (i=1; i<=n; ++i)
      zm[j] += z[i][j]; 
    zm[j] /= n;
  }
  mess (6, dvectorPrint (zm, pz, "zm"));

  for (j=1; j<=px; ++j) {
    wm[j] = 0.0;
    for (i=1; i<=n; ++i)
      wm[j] += (kn[i][1]*ww[i][j]);
    wm[j] /= vk;    
  }
  mess (6, dvectorPrint (wm, px, "wm"));


  /*** calculate szz - symmetric */

  for (i=1; i<=pz; ++i) {
    for (j=1; j<=i; ++j) {
      c = 0.0;
      for (k=1; k<=n; ++k)
        c += (z[k][i] - zm[i]) * (z[k][j] - zm[j]);
      szz[i][j] = c / (n-1);
      szz[j][i] = szz[i][j];
    }
  }
  mess (5, dmatrixPrint (r->szz, d->pz, d->pz, "szz"));

  /*** calculate sxz - non symmetric */

  for (i=1; i<=px; ++i) {
    for (j=1; j<=pz; ++j) {
      c = 0.0;
      for (k=1; k<=n; ++k)
        c += kn[k][1] * (ww[k][i] - wm[i]) * (z[k][j] - zm[j]);
      sxz[i][j] = c / v;
    }
  }
  mess (5, dmatrixPrint (r->sxz, d->px, d->pz, "sxz"));

  /*** calculate sxx - symmetric */

  for (i=1; i<=px; ++i) {
    for (j=1; j<=i; ++j) {
      c = 0.0;
      for (k=1; k<=n; ++k)
        c +=  kn[k][1] * (ww[k][i] - wm[i]) * (ww[k][j] - wm[j]);
      sxx[i][j] = (c-((n-1)*suu[i][j]))/v;
      sxx[j][i] = sxx[i][j];
    }
  } 
  mess (5, dmatrixPrint (r->sxx, d->px, d->px, "sxx"));

  /*** calculate E(X|Z,W) */

  S4 = (double ***) memory (kmax, sizeof (double ***));  S4--;

  for (k=1; k<=n; ++k) {

    /*** We need to create the s4 for every unique replication count (most
         data sets will have only a few). In order to avoid having to do 
         this every observation we keep track of our s4 matrices and only 
         create a new one as we encounter new replication sizes. */

    reps = d->kn[k][1];
    if (S4[reps] != NULL) {  
      s4 = S4[reps];
    }
    else { 

      s4 = dmatrix (px+pz, px+pz);
      S4[reps] = s4;

      /*** Setup and invert the symmetric s4 matrix. */

      for (i=1; i<=px; ++i) {   
        for (j=1; j<=i; ++j) {
          s4[i][j] = sxx[i][j]+suu[i][j]/reps; /* UL corner */
          s4[j][i] = s4[i][j];
        }
      }

      for (i=1; i<=px; ++i) {
        for (j=1; j<=pz; ++j) {
          s4[i][px+j] = sxz[i][j]; /* UR corner */
          s4[px+j][i] = sxz[i][j]; /* LL corner */
        }
      }

      for (i=1; i<=pz; ++i) {
        for (j=1; j<=pz; ++j) {
          s4[px+i][px+j] = szz[i][j]; /* LR corner */
        }
      }

      err = inverse (s4, px+pz);
      if (err) {
        q->status = StatusRCALSingular;
        if (!q->booting) 
          error (q->status, "Singular matrix in RCAL.");
        return;
      }

      note (6, "s4 for k=%d", reps);
      mess (6, dmatrixPrint (s4, px+pz, px+pz, "s4"));
    }


    /*** calculate the final result */

    for (j=1; j<=px; ++j) 
      wz[j]    = ww[k][j] - wm[j];
    for (j=1; j<=pz; ++j) 
      wz[px+j] = z[k][j] - zm[j];

    for (i=1; i<=px+pz; ++i) {
      wi[i] = 0.0;
      for (j=1; j<=px+pz; ++j)
        wi[i] += s4[i][j]*wz[j];
    }

    for (i=1; i<=px; ++i) {
      c = 0.0;
      for (j=1; j<=px; ++j) 
        c += sxx[i][j]*wi[j];
      for (j=1; j<=pz; ++j) 
        c += sxz[i][j]*wi[px+j];

      ex[k][i] = wm[i]+c;
    }
  }
  mess (5, dmatrixPrint (ex, d->n, d->px, "E[x]"));


  

  dnvectorFree (2, wz, wi);

  for (i=1; i<=kmax; ++i) {
    if (S4[i] != NULL) dmatrixFree (S4[i]);
  }
  freeMemory (++S4);

  q->status = StatusNoError;
}

/**********************************************************************/
/**********************************************************************/

/***  bootstrap function for regression calibration. */

void rcalBoot (dclData *d, dclQvf *q, dclRcal *r)
{
  rcal (d, q, r);

  if (q->status == StatusNoError) {
    makew (d, q, r->ex);
    qvf (d, q);
  }

  return;
}


/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*** Supporting routines calculate rcal covariance. */

void rcalVar_V (int i, double **V, dclData *d, dclRcal *r)
{
  int a, b, p, pw, pz, *ki;
  double k, v, n, sk, sk2;

  n  = d->n;     
  p  = d->p;     
  pw = d->px;    
  pz = d->pz;    
  ki = d->kn[i]; 

  n = (double) d->n;        
  k = (double) d->kn[i][1]; 
  v = r->v;                 

  sk  = pw*k;
  sk2 = pw*k*k;

  dmatrixZero (V, p, p);
  for (a=1; a<=pw; ++a) { 
    for (b=1; b<=pw; ++b)
      V[1+a][1+b] = r->sxx[a][b]* (n*k/v)*(1-(2*k/sk)+(sk2/(sk*sk))) +
      r->suu[a][b]* (n/v)*(1-k/sk);                 
    for (b=1; b<=pz; ++b) {
      V[1+a][1+pw+b] = r->sxz[a][b] * (n/v)*(k-(k*k/sk));         
      V[1+pw+b][1+a] = V[1+a][1+pw+b];                            
    }
  }
  for (a=1; a<=pz; ++a) {
    for (b=1; b<=pz; ++b)
      V[1+pw+a][1+pw+b] = r->szz[a][b]; 
  }

  note (6, "V[i] n=%d, k=%d (reps), nu=%f", (int) n, (int) k, v);
  mess (6, dmatrixPrint (V, p, p, "V"));  
}


void rcalVar_Psi1 (int i, double **Psi1, double **V, double *rw, double *rz, 
                   dclData *d, dclRcal *r)
{
  int a, b, p, pw, pz;
  double n, v, k;

  p  = d->p;     
  pw = d->px;    
  pz = d->pz;    

  

  for (a=1; a<=pw; ++a) {
    rw[a] = d->x[i][a] - r->wm[a];
  }
  for (a=1; a<=pz; ++a) {
    rz[a] = d->z[i][a] - r->zm[a];
  }
  mess (7, dvectorPrint (rw, pw, "rw"));  
  mess (7, dvectorPrint (rz, pz, "rz"));  

  

  n = (double) d->n;        
  k = (double) d->kn[i][1]; 
  v = r->v;                 

  dmatrixZero (Psi1, p, p);
  for (a=1; a<=pw; ++a) { 
    for (b=1; b<=pw; ++b) 
      Psi1[1+a][1+b] = (n*k/v) * rw[a]*rw[b];    
    for (b=1; b<=pz; ++b) {
      Psi1[1+a][1+pw+b] = (n*k/v) * rw[a]*rz[b]; 
      Psi1[1+pw+b][1+a] = Psi1[1+a][1+pw+b]; 
    }
  }
  for (a=1; a<=pz; ++a) {
    for (b=1; b<=pz; ++b)
      Psi1[1+pw+a][1+pw+b] = (n/(n-1)) * rz[a]*rz[b]; 
  }
  mess (7, dmatrixPrint (Psi1, p, p, "Psi1*"));  

  

  for (a=1; a<=p; ++a) {
    for (b=1; b<=p; ++b)
      Psi1[a][b] -= V[a][b];
  }

  mess (7, dmatrixPrint (Psi1, p, p, "Psi1"));
}


void rcalVar_Psi2 (int i, double **Psi2, double kmean, double **T,  double *t,
                   dclData *d, dclRcal *r)
{
  int a, b, j, k, n, p, pw, pz;

  if (r->uuignore == True) return;  

  n  = d->n;     
  p  = d->p;     
  pw = d->px;    
  pz = d->pz;    

  dmatrixZero (Psi2, p, p);  
  if (kmean == 0.0) return;

  dmatrixZero (T, pw, pw);

  k = d->kn[i][1];        
  for (j=1; j<=k; ++j) {  
    for (a=1; a<=pw; ++a) {
      
      t[a] = d->W[i][(a-1)*k+j] - d->x[i][a]; 
    }

    for (a=1; a<=pw; ++a) {
      for (b=1; b<=pw; ++b) 
        T[a][b] += t[a]*t[b];
    }
  }

  for (a=1; a<=pw; ++a) {
    for (b=1; b<=pw; ++b) 
      Psi2[1+a][1+b] = (T[a][b] - ((k-1)*r->suu[a][b])) / kmean;
  }

  mess (7, dmatrixPrint (Psi2, p, p, "Psi2"));
}


void rcalVar_DS (double **D, double **S, double **DS, double **T1, 
                 double *t1, dclData *d, dclRcal *r)
{
  int a, b, p, pw, pz;

  p  = d->p;
  pw = d->px;
  pz = d->pz;

  dmatrixZero (D, p, p);

  D[1][1] = 1.0;
  for (a=1; a<=pw; ++a) {
    for (b=1; b<=pw; ++b)
      D[1+a][1+b] = r->sxx[a][b] + r->alpha*r->suu[a][b]; 
    for (b=1; b<=pz; ++b) {
      D[1+a][1+pw+b] = r->sxz[a][b]; 
      D[1+pw+b][1+a] = r->sxz[a][b]; 
    }
  }
  for (a=1; a<=pz; ++a) {
    for (b=1; b<=pz; ++b)
      D[1+pw+a][1+pw+b] = r->szz[a][b]; 
  }
  mess (6, dmatrixPrint (D, p, p, "D")); 

  dmatrixZero (S, p, p);
  for (a=1; a<=pw; ++a) { 
    for (b=1; b<=pw; ++b)
      S[1+a][1+b] = r->suu[a][b];
  }
  mess (6, dmatrixPrint (S, p, p, "S"));

  for (a=1; a<=p; ++a) {
    for (b=1; b<=p; ++b)
      DS[a][b] = D[a][b] - (r->alpha*S[a][b]);  
  }
  mess (6, dmatrixPrint (DS, p, p, "D-alpha*S"));
}


void rcalVar_Ci (int i, double **C, double **D, double **S, 
                 dclData *d, dclQvf *q, dclRcal *r)
{
  int a, b, err;
  double k;

  k = (double) d->kn[i][1];

  for (a=1; a<=d->p; ++a) {
    for (b=1; b<=d->p; ++b)
      C[a][b] = D[a][b] - (r->alpha-(1/k))*S[a][b];
  }

  note (6, "C[i] for reps=%d", (int) k);
  mess (6, dmatrixPrint (C, d->p, d->p, "Ci (not inverted)"));

  err = inverse (C, d->p);
  if (err) {
    q->status = StatusRcalVarSingular;
    error (q->status, "Singular matrix in Rcal Variance (Ci).");
  }  

  mess (6, dmatrixPrint (C, d->p, d->p, "C[i]"));
}


void rcalVar_Ri (int i, double *R, dclData *d, dclRcal *r)
{
  int k;

  R[1] = 1;

#if 1 

  for (k=1; k<=d->px; ++k)  
    R[1+k] = d->x[i][k];

  for (k=1; k<=d->pz; ++k)
    R[1+d->px+k] = d->z[i][k];

#else

  for (k=1; k<=d->px; ++k)  
    R[1+k] = d->x[i][k] - r->wm[k];

  for (k=1; k<=d->pz; ++k)
    R[1+d->px+k] = d->z[i][k] - r->zm[k];

#endif

  mess (7, dvectorPrint (R, d->p, "R[i]"));
}


void rcalVar_Qi (int i, double *Q, double **C, double **DS, double *R, 
                 double **T1, dclData *d, dclRcal *r)
{
  int a, b, p;

  p = d->p;

  matrix_AB (T1, DS, p, p, C, p);  

  for (a=1; a<=p; ++a) {
    Q[a] = 0;
    for (b=1; b<=p; ++b)
      Q[a] += T1[a][b]*R[b];       
  }

  mess (7, dvectorPrint (Q, p, "Q[i]"));
}



void rcalVar_P (int i, double *P1, double *P2, double **C, double **DS, 
                double *beta, double **I, double **T1, double **T2, 
                dclData *d, dclRcal *r)
{
  int a, b, p;
  double k;

  p = d->p;

  note (6, "P1 & P2 for reps=%d", d->kn[i][1]);

  matrix_AB (T1, C, p, p, DS, p);    
  mess (6, dmatrixPrint (T1, p, p, "Cj*(D-alpha*S)"));

  /* P1 (used for d1) */

  for (a=1; a<=p; ++a) {
    for (b=1; b<=p; ++b)
      T2[a][b] = I[a][b] - T1[a][b]; 
  }
  mess (6, dmatrixPrint (T2, p, p, "T1=I-Cj*{D-(alpha*S)}"));

  for (a=1; a<=p; ++a) {
    P1[a] = 0.0;
    for (b=1; b<=p; ++b)
      P1[a] += T2[a][b] * beta[b];  
  }
  mess (6, dvectorPrint (P1, p, "P1=T1*beta"));

  /* P2 (use for d2) */

  matrix_AB (T1, DS, p, p, C, p);    
  mess (6, dmatrixPrint (T1, p, p, "(D-alpha*S)*Cj"));

  k = (double) d->kn[i][1];
  for (a=1; a<=p; ++a) {
    for (b=1; b<=p; ++b)
      T2[a][b] = (r->alpha-(1/k)) * T1[a][b] - r->alpha*I[a][b]; 
  }
  mess (6, dmatrixPrint (T2, p, p, "T2=(alpha-1/k[i])*{D-(alpha*S)}*Cj-alpha*I"));

  for (a=1; a<=p; ++a) {
    P2[a] = 0.0;
    for (b=1; b<=p; ++b)
      P2[a] += T2[a][b] * beta[b];  
  }
  mess (6, dvectorPrint (P2, p, "P2=T2*beta"));
}


double rcalVar_q (int p, int p1, double *R, double **C, double **Psi, double *P, 
                  double *t1, double *t2)
{
  int a, b;
  double q;

  /*** Routine to calculate scalar q = R*C*Psi*P where R & P are 
       1xp vectors and C & Psi are pxp matrices. The vector elements 
       pertain to B=[1 W Z]  where W is 1xpw, Z is 1xpz and p=1+pw+pz. 
       The matrix elements pertain to B'B. */

  /*** This is the most time intensive routine in rcalVar() partly
       because it is called 2*n^2 times. As such every optimization 
       counts! When I began over 90% of the time was spent in this 
       routine, after re-coding and optimization the time spend 
       here is significantly less. */

  /*** This is the final version of our algorithm. We greatly improved
       speed using the following optimizations:
       
       1) We have reduced the number of multiplications from the 
          initial 2*p^3+2*p^2+p to 2*p^2+p. We accomplished this 
          by precomputing one matrix by vector operation outside of 
          the loop (P_px1 = DSC_pxp * beta_px1) and by rearranging 
          the remaining algebra to favor matrix by vector operations 
          (p^2) instead of matrix by matrix (p^3). This resulted in 
          a 5 to 10 fold speed improvement.
          
       2) We made use of the special structure of the matrices to 
          reduce the number of operations even further. 
          
          a) The intercept (row & column 1) does not contribute 
             to q and can be ignored. As such, all our loops start 
             with 2 instead of 1.

          b) When calculating q using Psi2 we only need concern 
             ourselves with those elements that involve W. Thus we
             loop from 2 to p1=1+pw instead of 2 to p=1+pw+pz. 
             When calculating q with Psi1 we make p1=p.

          In one test with 4 variables this these reduction in
          operations resulted in a further 45% speed improvement 
          in rcalVar(). */

#define ISTART 2  

  for (a=ISTART; a<=p1; ++a) {
    t1[a] = 0.0;
    for (b=ISTART; b<=p; ++b)
      t1[a] += R[b] * C[b][a];     
  }

  for (a=ISTART; a<=p1; ++a) {
    t2[a] = 0.0;
    for (b=ISTART; b<=p1; ++b)
      t2[a] += t1[b] * Psi[b][a];  
  }

  q = 0.0;
  for (a=ISTART; a<=p1; ++a) {
    q += t2[a] * P[a];             
  }

#undef ISTART

  return (q);
}


/*** Flip the final p x p matrices from [1 W Z] to [1 Z W] order. */

void rcalMatrixFlip (double **B, double **T, int p, int pw, int pz)
{
  int a, b;

#if 0  
  int c=1;
  for (a=1; a<=p; ++a) 
    for (b=1; b<=a; ++b, ++c) {
      B[a][b] = B[b][a] = c;
      T[a][b] = T[b][a] = 0;
    }
#endif

    

    T[1][1] = B[1][1];
    for (a=1; a<=pz; ++a) {
      T[1][1+a] = B[1][1+pw+a];
      T[1+a][1] = B[1+pw+a][1];
      for (b=1; b<=pz; ++b)
        T[1+a][1+b] = B[1+pw+a][1+pw+b];
      for (b=1; b<=pw; ++b) {
        T[1+pz+b][1+a] = B[1+b][1+pw+a];
        T[1+a][1+pz+b] = B[1+pw+a][1+b];
      }
    }

    for (a=1; a<=pw; ++a) {
      T[1][1+pz+a] = B[1][1+a];
      T[1+pz+a][1] = B[1+a][1];
      for (b=1; b<=pw; ++b)
        T[1+pz+a][1+pz+b] = B[1+a][1+b];;
    }

    dmatrixCopy (B, T, p, p);

    
}


/*** Calculate the Rcal covariance estimator (sandwich & asymptotic). 
     This code implements the standard errors as described on Page 72 
     of "Measurement Error in Nonlinear Models" by R.J. Carroll, 
     D. Rupert and L.A. Stefanski. */

double theTime (void);      
char  *stringTime (double); 

void rcalVar (int type, dclData *d, dclQvf *q, dclRcal *r)  
{
  int i, j, k, n, p, pw, pz, a, b, err, poll=True, kmax;
  double kmean, ksum, q1, q2, time, start=0;
  double **D, **S, **Psi1, **Psi2, **I, **DS,  
         **A0, **B1, **B2, **T1, **T2;         
  double *beta, *d1, *d2, *ri, *ei, *t1, *t2;  
  double ***C, ***V, **P1, **P2;  
  double **R, **Q, *s;  


  note (6, "*** Rcal: Calculating Rcal covariance");
  note (10, "qvfVar()");

  n  = d->n;  
  p  = d->p;  
  pw = d->px; 
  pz = d->pz; 

  dnvector (7, p, &d1, &d2, &ri, &ei, &t1, &t2, &beta); 
  dnmatrix (11, p, p, &D, &S, &DS, &Psi1, &Psi2, &I,    
                      &B1, &B2, &T1, &T2, &A0);
  dnmatrix (2, n, p, &R, &Q);  
  s = dvector (n);

  /************************************************************************/

  /*** Note: q->var, q->eta, q->mu, q->v need to be calculated
             with prior call to qvfVariance(). */

  dInvLink (d, q, q->d1, q->mu, q->eta);
  mess (6, dvectorPrint (q->d1,  d->n, "1st derivative inverse link"));

  
  
  

  /************************************************************************/

  /*** Note: The algorithm presupposes beta = [intercept, W, Z] (measurement
             error parameters are before parameters measured without error).
             This is different from our default so we need to flip our beta's
             and adjust our results (B1 & B2) for this later. */

  beta[1] = q->beta[1];           
  for (i=1; i<=pw; ++i) 
    beta[1+i] = q->beta[1+pz+i];  
  for (i=1; i<=pz; ++i) 
    beta[1+pw+i] = q->beta[1+i];  

  /************************************************************************/

  /*** We calculate kmean (which we need to calculate Psi2) and kmax (the 
       largest number of measurement error replications in our data set). 
       We use kmax to help us determine the size of the data structures to 
       hold some of the precomputed matrices and vectors that change only 
       based on the number of replications. */

  ksum = 0;
  kmax = -1;
  for (i=1; i<=d->n; ++i) { 
    k = d->kn[i][1];
    ksum += k-1;
    if (k > kmax) kmax = k;
  }
  kmean = ksum/n;
  note (6, "mean sum (k[i]-1) = %f  max k[i] = %d", kmean, kmax);

  


  /************************************************************************/

  /*** We form some static constants & matrices I, D, S & DS that we need. */

  dmatrixZero (I, p, p);
  for (a=1; a<=p; ++a) 
    I[a][a] = 1.0;

  r->alpha = (n-1) / r->v;
  note (6, "n=%d, nu=%f, alpha = %f", n, r->v, r->alpha);

  rcalVar_DS (D, S, DS, T1, t1, d, r);   


  /*** We precompute as much as possible. Our intermediary matrices 
       C, V, P1 & P2 of size p x p depend only on the number of measurement
       error replicates. Eg. C[k] would be the C matrix for any observation 
       with k replicates. */

  C  = (double ***) memory (kmax, sizeof (double ***));  C--;
  V  = (double ***) memory (kmax, sizeof (double ***));  V--;
  P1 = (double **)  memory (kmax, sizeof (double **));   P1--;
  P2 = (double **)  memory (kmax, sizeof (double **));   P2--;

  for (i=1; i<=n; ++i) {
    k = d->kn[i][1];
    if (C[k] == NULL) {                   
      dnmatrix (2, p, p, &C[k], &V[k]);   
      dnvector (2, p, &P1[k], &P2[k]);

      rcalVar_Ci (i, C[k], D, S, d, q, r); 
      rcalVar_V  (i, V[k], d, r);          
      rcalVar_P  (i, P1[k], P2[k], C[k], DS, beta, I, T1, T2, d, r); 
    }
  }


  /*** Precompute vectors Ri, Qi of size p for every observation. Also 
       calculate the vector s[i] of size n (observations). We use a lot 
       of extra storage for large n, but we only calculate these values
       n times instead of n^2. */

  for (i=1; i<=n; ++i) {
    k = d->kn[i][1];
    rcalVar_Ri (i, R[i], d, r); 
    rcalVar_Qi (i, Q[i], C[k], DS, R[i], T1, d, r);
    s[i] = (q->d1[i]*q->d1[i])/q->v[i];
    
    
  }


  /************************************************************************/

  /*** Our time intensive n by n loop. This double loop can be CPU intensive 
       enough that we need to check for windowing events inside the inner loop.
       We will also estimate the time to completion. At the end we will have 
       our A0, B1 & B2 matrices which we use to calculate the covariance. */

  resetWinMessages();

  dmatrixZero (B1, p, p);  
  dmatrixZero (B2, p, p);
  dmatrixZero (A0, p, p);

  for (i=1; i<=n; ++i) {

    /* Since for large data sets this loop can take a *very* long 
       time, we will print out an estimated time to completion. */

    if (i == 1) {
      start = theTime();            
      poll = True;                  
    }
    else if (poll) {
      time = theTime() - start;     
      if (time >= 5000) {           
        time = (time/(i-1))*n/1000; 
        if ( time > 30)             
          note (2, "Estimated time to calculate rcal variance is %s.", 
                   stringTime(time));
        poll = False;               
      }
    }

    /* End timing section. Calculate Psi1, Psi2 & ri for this observation. */

    k = d->kn[i][1];

    rcalVar_Psi1 (i, Psi1, V[k], t1, t2, d, r);
    rcalVar_Psi2 (i, Psi2, kmean, T1, t1, d, r);

    for (k=1; k<=p; ++k) {
      ri[k] = ((d->y[i]-q->mu[i]) * q->d1[i] / (q->v[i])) * Q[i][k];
    }

    /*** Inner j loop */

    dvectorZero (d1, p);  
    dvectorZero (d2, p);

    for (j=1; j<=n; ++j) {

      if (j%100 == 0) handleWinMessages(WINLOOP);

      k = d->kn[j][1];

      q1 = rcalVar_q (p, p, R[j], C[k], Psi1, P1[k], t1, t2);
      for (a=1; a<=p; ++a) {
        d1[a] += s[j]*Q[j][a]*q1;
      }

      if (r->uuignore != True) {
        q2 = rcalVar_q (p, 1+pw, R[j], C[k], Psi2, P2[k], t1, t2);
        for (a=1; a<=p; ++a) {
          d2[a] += s[j]*Q[j][a]*q2;
        }
      }

    } 

    for (a=1; a<=p; ++a) {
      ei[a] = ri[a] - d1[a]/n - d2[a]/n;
    }

    for (a=1; a<=p; ++a) {
      for (b=1; b<=p; ++b) {
        B1[a][b] += ri[a]*ri[b];  
        B2[a][b] += ei[a]*ei[b];  
        A0[a][b] += Q[i][a]*Q[i][b]*s[i];
      }
    }

  } 

  matrix_Ac (B1, B1, 1.0/n, p, p);
  matrix_Ac (B2, B2, 1.0/n, p, p);
  matrix_Ac (A0, A0, 1.0/n, p, p);


  /*** We have our A0, B1 & B2 matrices. We rearrange their structure to
       conform to our naive covariance matrix q->cov. We then proceeed to 
       calculate our asymptotic or sandwich variance. */

  rcalMatrixFlip (B1, T1, p, pw, pz); 
  rcalMatrixFlip (B2, T1, p, pw, pz); 
  rcalMatrixFlip (A0, T1, p, pw, pz); 

  mess (5, dmatrixPrint (B1, p, p, "B1"));
  mess (5, dmatrixPrint (B2, p, p, "B2"));
  mess (5, dmatrixPrint (A0, p, p, "A0"));

  if (type == CovRcal) {

    

    for (a=1; a<=p; ++a) {
      for (b=1; b<=p; ++b) 
        B2[a][b] = A0[a][b]*q->var + B2[a][b] - B1[a][b];
    }
    mess (5, dmatrixPrint (q->cov, p, p, "B2,i"));

  } 

  err = inverse(A0, p);
  if (err != 0) {
    q->status = StatusRcalVarSingular;
    error (q->status, "Singular matrix in Rcal (Sandwich).");
  }
  mess (6, dmatrixPrint (A0, p, p, "Inv(A0)"));  

  matrix_ABtC (q->cov, T1, A0, p, p, B2, p, A0, p);
  matrix_Ac   (q->cov, q->cov, 1.0/n, p, p);  

  mess (5, dmatrixPrint (q->cov, p, p, "rcal cov[] = 1/n*A0(B2)A0"));


  /************************************************************************/

  /*** Free our workspace & exit. */

  dnvectorFree (8, d1, d2, ri, ei, t1, t2, beta, s);
  dnmatrixFree (13, D, DS, S, Psi1, Psi2, I, B1, B2, T1, T2, R, Q, A0);

  for (i=1; i<=kmax; ++i) {
    if (C[i] != NULL) {
      dnmatrixFree (2, C[i], V[i]);
      dnvectorFree (2, P1[i], P2[i]);
    }
  }
  freeMemory (++C);
  freeMemory (++V);
  freeMemory (++P1);
  freeMemory (++P2);
}
